from balder.fields.offsetfiltered import BalderFilteredWithOffset
from balder.fields.filtered import BalderFiltered
from balder.types.query.meta import BalderQueryMeta
from balder.types.utils import classToString
import graphene
import django_filters
from graphene_django import DjangoObjectType
from graphene.utils.props import props


class IllConfigured(Exception):
    pass

class NoResolveError(Exception):
    pass



class BalderQuery(metaclass=BalderQueryMeta):

    class Meta:
        abstract = True
        list = None
        filter = None
        operation = None


    @classmethod
    def cls_resolve(cls, root, info, *args, **kwargs):
        try:
            return cls.resolve(root, info, *args, **kwargs)
        except NoResolveError as e:
            raise NoResolveError(f"Please overwride resolve in {cls.__name__} at {cls.__module__}")

    @classmethod
    def _get_description(cls) -> str:
        return cls.__doc__

    @classmethod
    def _get_resolver_or_none(cls) -> str:
        if hasattr(cls, "resolve"):
            return cls.resolve
        else:
            return None

    @classmethod
    def _get_operation(cls) -> str:
        """Gets the opereration name (string literal in query)

        Returns:
            str: the operation name
        """
        if cls._meta.operation is not None: return cls._meta.operation
        return classToString(cls)

    @classmethod
    def _to_field(cls) -> graphene.Field:
        meta = cls._meta
        assert meta.type is not None, f"Please provide a Meta class with at least the attribute type in {cls.__name__}"

        resolver = cls._get_resolver_or_none()
        if resolver:
            # Already provided resolver overwrites every default
            assert cls._arguments is not None, f"If you are using a custom resolver please also provide an Arguments class (can pass) in {cls.__name__}"
            attributes = props(cls._arguments)
            if meta.list:
                return graphene.List(meta.type, resolver=resolver, description=cls._get_description(), **attributes)
            else:
                return graphene.Field(meta.type, resolver=resolver, description=cls._get_description(), **attributes)

        assert issubclass(meta.type, DjangoObjectType), "You cannot use Query without a custom resolver if it is not a DjangoObjectType"
        qs = meta.queryset or meta.type._meta.model._default_manager.get_queryset()

        if meta.personal:
            # Personal means we want to filter the queryset for the user with that fied
            assert isinstance(meta.personal, str), "Meta attribute personal must be of type str"
            list_resolver = lambda root, info, *args, **kwargs: qs.filter(**{meta.personal: info.context.user})
            item_resolver = lambda root, info, *args, **kwargs: qs.filter(**{meta.personal: info.context.user}).first()
        else:
            list_resolver = lambda root, info, *args, **kwargs: qs.all()
            item_resolver = lambda root, info, *args, **kwargs: qs.first()

        if meta.filter:
            if issubclass(meta.filter, django_filters.FilterSet):
                if meta.paginate:
                    return BalderFilteredWithOffset(meta.type,  description=cls._get_description(), filterset_class=meta.filter, queryset_resolver=list_resolver)
                
                return BalderFiltered(meta.type,  description=cls._get_description(), filterset_class=meta.filter, queryset_resolver=list_resolver)
            else:
                raise IllConfigured("Meta attribute filter must be of type django_filters.FilterSet")
        
        if meta.list:
            return graphene.List(meta.type, resolver=list_resolver, description=cls._get_description())
        else:
            return graphene.Field(meta.type, resolver=item_resolver, description=cls._get_description())



    


import graphene
from graphene_django.converter import convert_django_field
from django.db import models
from graphene.types.generic import GenericScalar
from balder.enum import InputEnum


@convert_django_field.register(models.JSONField)
def convert_json_field_to_string(field, registry=None):
    return GenericScalar()


class BalderRegistry:

    def __init__(self) -> None:
        
        self.queries = {"void": graphene.String(default_value="Hallo")}
        self.mutations = {}
        self.subscriptions = {}
        self.types = []

    def registerType(self, type):
        self.types.append(type)


    def registerQuery(self, query):
        self.queries[query._get_operation()] = query._to_field()

    def registerMutation(self, mutation):
        self.mutations[mutation._get_operation()] = mutation._to_field()

    def registerSubscription(self, subscription):
        self.subscriptions[subscription._get_operation()] = subscription._to_field()
    

    def buildSchema(self, query=None, mutation=None, subscription=None, types=[]):
        assert issubclass(query, graphene.ObjectType) or query is None, "If you provide an additional root Query please make sure its of type Query"
        QueryBase = query if query is not None else graphene.ObjectType
        MutationBase = mutation if mutation is not None else graphene.ObjectType
        SubscriptionBase = subscription if subscription is not None else graphene.ObjectType

        query = type("Query", (QueryBase, ), {**self.queries, "__doc__": "The root Query"}) if self.queries != {} else None
        mutation = type("Mutation", (MutationBase, ), {**self.mutations, "__doc__": "The root Mutation"}) if self.mutations != {} else None
        subscription = type("Subscription", (SubscriptionBase, ), {**self.subscriptions, "__doc__": "The root Subscriptions"}) if self.subscriptions != {} else None

        print(self.types)

        return graphene.Schema(
            query = query,
            mutation = mutation,
            subscription = subscription,
            types = self.types + types
        )









BALDER_REGISTRY = None

def get_balder_registry():
    global BALDER_REGISTRY
    if BALDER_REGISTRY is None:
        BALDER_REGISTRY = BalderRegistry()
    return BALDER_REGISTRY


def register_type(cls):
    registry = get_balder_registry()
    if cls not in registry.types:
        registry.registerType(cls)
    return cls


from functools import partial
from graphene import Field, List
from graphene.types.utils import get_type
from graphene_django.filter.utils import (
    get_filtering_args_from_filterset,
    get_filterset_class,
)


class BalderFiltered(Field):
    """[summary]

    Args:
        Field ([type]): [description]
    """

    def __init__(
        self,
        _type,
        related_field=None,
        fields=None,
        extra_filter_meta=None,
        filterset_class=None,
        queryset_resolver=None,
        *args,
        **kwargs
    ):
        """A Filtered field for any Query

        Args:
            _type ([type]): [description]
            related_field ([type], optional): The related field that we are using the manager from. Defaults to None.
            fields ([type], optional): [description]. Defaults to None.
            extra_filter_meta ([type], optional): [description]. Defaults to None.
            filterset_class ([type], optional): [description]. Defaults to None.
            queryset_resolver ([type], optional): Function that resolves the initial queryset for the filter, is overwritten by related_field. Default to None
        """
        _type = get_type(_type)
        _fields = _type._meta.filter_fields
        _model = _type._meta.model

        standard_resolver = (
            lambda root, info, *args, **kwargs: _model._default_manager.get_queryset()
        )
        self._relatedfield = related_field
        self._queryset_resolver = queryset_resolver or standard_resolver
        self.of_type = _type

        self.fields = fields or _fields
        meta = dict(model=_model, fields=self.fields)
        if extra_filter_meta:
            meta.update(extra_filter_meta)
        self.filterset_class = get_filterset_class(
            filterset_class, **meta
        )  # create one if there is no filtersetclass provided
        self.filtering_args = get_filtering_args_from_filterset(
            self.filterset_class, _type
        )
        kwargs.setdefault("args", {})
        kwargs["args"].update(self.filtering_args)
        super().__init__(List(_type), *args, **kwargs)

    @staticmethod
    def related_resolver(
        related_field, filterset_class, filtering_args, root, info, *args, **kwargs
    ):
        # TODO: here it would be good to filter for permissions if it is ndeeded

        filter_kwargs = {k: v for k, v in kwargs.items() if k in filtering_args}
        qs = getattr(root, related_field)
        qs = filterset_class(data=filter_kwargs, queryset=qs).qs
        return qs

    @staticmethod
    def list_resolver(
        queryset_resolver, filterset_class, filtering_args, root, info, *args, **kwargs
    ):
        # TODO: here it would be good to filter for permissions if it is ndeeded

        filter_kwargs = {k: v for k, v in kwargs.items() if k in filtering_args}
        qs = filterset_class(
            data=filter_kwargs, queryset=queryset_resolver(root, info, *args, **kwargs)
        ).qs
        return qs

    def get_resolver(self, parent_resolver):
        if self._relatedfield:
            # if related field is set we are dealing with a related field manager
            # so we will get the manager from the root object (through accessing the related field)
            return partial(
                self.related_resolver,
                self._relatedfield,
                self.filterset_class,
                self.filtering_args,
            )
        else:
            # We are just using the default manager of this class...
            # TODO: Maybe warn if we are doing this from a related field
            return partial(
                self.list_resolver,
                self._queryset_resolver,
                self.filterset_class,
                self.filtering_args,
            )



import logging


from balder.schema import graphql_schema
logger = logging.getLogger(__name__)


try:
    import channels_graphql_ws
    from channels.generic.websocket import AsyncJsonWebsocketConsumer

    class MyGraphqlWsConsumer(channels_graphql_ws.GraphqlWsConsumer):
        """Channels WebSocket consumer which provides GraphQL API."""
        schema = graphql_schema
        # Uncomment to send keepalive message every 42 seconds.
        send_keepalive_every = 15

        # Uncomment to process requests sequentially (useful for tests).
        # strict_ordering = True

        async def on_connect(self, payload):
            """New client connection handler."""
            # You can `raise` from here to reject the connection.
            user = self.scope.get('user')
            logger.debug(f"New client connected with user {self.scope.get('user')}")

except ImportError as e:

    MyGraphqlWsConsumer = None



from balder.registry import get_balder_registry
import graphene
from balder.autodiscover import autodiscover

class Query(graphene.ObjectType):
    hello = graphene.String(default_value="Hi!")

    def resolve_hello(*args, **kwargs):
        return "Hallo"


# Autodiscover for all of the Balder Modules in the installed Apps

autodiscover()

graphql_schema = get_balder_registry().buildSchema(query = Query)


from balder.types.mutation.meta import BalderMutationMeta
from balder.types.utils import classToString
import graphene

class IllConfigured(Exception):
    pass

class NoMutateException(Exception):
    pass



class BalderMutation(metaclass=BalderMutationMeta):
    Arguments = None # Should be overwriten by the actual mutation


    class Meta:
        abstract = True


    @classmethod
    def _get_description(cls) -> str:
        return cls.__doc__

    @classmethod
    def _get_mutate_or_none(cls) -> str:
        if hasattr(cls, "mutate"):
            return cls.mutate
        else:
            return None

    @classmethod
    def _get_operation(cls) -> str:
        """Gets the opereration name (string literal in query)

        Returns:
            str: the operation name
        """
        if cls._meta.operation is not None: return cls._meta.operation
        return classToString(cls)

    @classmethod
    def _to_field(cls) -> graphene.Field:
        meta = cls._meta
        assert meta.type is not None, f"Please provide a Meta class with at least the attribute type in {cls.__name__}"
        assert cls.Arguments is not None, f'Please provide a Arguments class (can just pass in body if no arguments are needed) in  {cls.__name__}'

        mutate_function = cls._get_mutate_or_none()
        assert mutate_function is not None, f"Please provide a mutate function in  {cls.__name__}"

        mutation = type("Mutation", (graphene.Mutation,), {"Arguments": cls.Arguments, "Output": meta.type, "mutate": mutate_function, "__doc__": cls._get_description()})
        return mutation.Field(description=cls._get_description())




    


from graphene_django import DjangoObjectType


class BalderObject(DjangoObjectType):

    class Meta:
        abstract= True


    def resolve_tags(self, *args, **kwargs):
        if isinstance(self.tags, list):
            return map(str, self.tags)
        else:
            return map(str, self.tags.all())